"""
Base data types which are extended with concrete OpenId data types in :mod:`simple_openid_connect.data`
"""
import abc
import logging
from typing import Any, List, Literal, Type, TypeVar

from cryptojwt import JWK, JWS
from furl import Query, furl
from pydantic import BaseModel

logger = logging.getLogger(__name__)


Self = TypeVar("Self", bound="OpenidBaseModel")


class OpenidBaseModel(BaseModel, metaclass=abc.ABCMeta):
    """
    Base model type upon which all openid data types are built.

    It implements encoding and decoding functionality that are commonly used in OpenID contexts.
    """

    def encode_x_www_form_urlencoded(self) -> str:
        """
        Encode this message as a `x-www-form-urlencoded` formatted string.

        This is useful to send the message to an OP (if it is a request) either directly as GET url parameters or as an
        `x-www-form-urlencoded` request body

        """
        query = Query()
        query.set(self.dict(exclude_defaults=True))
        return query.encode()  # type: ignore # because furl has no typedefs, but we know what this returns

    def encode_url(self, url: str) -> str:
        """
        Encode this message as query string parameters into the existing url.

        This method explicitly only encodes the message into an urls query string because Openid specifies that only
        responses can be returned via a fragment and since this library is only intended for usage as a relying party,
        it should never need to generate responses.
        """
        url_parsed = furl(url)
        url_parsed.args.update(self.dict(exclude_defaults=True))
        return url_parsed.tostr()  # type: ignore # because furl has no typedefs, but we know what this returns

    @classmethod
    def parse_x_www_form_urlencoded(cls: Type[Self], s: str) -> Self:
        """
        Parse a received message that is parsed from the given `x-www-form-urlencoded` formatted string.
        """
        query = Query(s)
        one_value_params = {key: query.params[key] for key in query.params.keys()}
        return cls.parse_obj(one_value_params)

    @classmethod
    def parse_url(
        cls: Type[Self],
        url: str,
        location: Literal["query", "fragment", "auto"] = "auto",
    ) -> Self:
        """
        Parse a received message that is encoded as part of the URL as query parameters.

        :param url: The url which contains a message either in its query string or fragment
        :param location: Where the message data is located in the url.
            If set to 'auto', fragment will be tried first with query being used as a fallback.
        """
        if location == "query":
            return cls.parse_x_www_form_urlencoded(str(furl(url).query))
        elif location == "fragment":
            fragment = furl(url).fragment.query
            return cls.parse_x_www_form_urlencoded(str(fragment))
        elif location == "auto":
            try:
                return cls.parse_url(url, location="fragment")
            except Exception as e:
                logger.debug(
                    "Could not parse %s from fragment, trying query string: %s",
                    cls.__name__,
                    e,
                )
                return cls.parse_url(url, location="query")
        else:
            raise ValueError(f"invalid location value {location}")

    @classmethod
    def parse_jwt(cls: Type[Self], value: str, signing_keys: List[JWK]) -> Self:
        """
        Parse received data that is encoded as a signed Json-Web-Token (JWT).

        :param value: The encoded JWT
        :param signing_keys: List of keys one of which has been used to sign the JWT
        """
        verifier = JWS()
        msg = verifier.verify_compact(value, signing_keys)
        return cls.parse_obj(msg)



import pytest
import requests
from django.shortcuts import resolve_url
from requests.exceptions import ConnectionError


def is_provider_running() -> bool:
    try:
        requests.get("http://localhost:3000")
        return True
    except ConnectionError:
        return False


@pytest.mark.skipif(not is_provider_running(), reason="dummy provider not running")
def test_directly_calling_login_endpoint(dummy_ua, live_server):
    # initiate login
    init_response = dummy_ua.naviagte_to(
        live_server + resolve_url("simple_openid_connect_django:login")
    )
    assert init_response.ok

    # perform login and give consent on the provider
    login_callback = dummy_ua.login_to_dummy_provider(init_response.url)

    # assert the page can now be accessed
    login_response = dummy_ua.naviagte_to(live_server + login_callback)
    assert login_response.ok
    assert login_response.content == b"default login redirect view"


@pytest.mark.skipif(not is_provider_running(), reason="dummy provider not running")
def test_directly_accessing_protected_resource(dummy_ua, live_server):
    # initiate login by accessing a protected resource
    init_response = dummy_ua.naviagte_to(
        live_server + resolve_url("test-protected-view")
    )
    assert init_response.ok

    # perform login and give consent on the provider
    login_callback = dummy_ua.login_to_dummy_provider(init_response.url)
    login_response = dummy_ua.naviagte_to(live_server + login_callback)

    # assert
    assert login_response.ok
    assert login_response.content == b"hello user test"



import logging
import random
import string
from pathlib import Path
from typing import Dict, List, Mapping, Tuple

import pytest
import requests
import responses
import yaml
from furl import furl
from requests import PreparedRequest
from requests.adapters import HTTPAdapter

from simple_openid_connect.data import (
    AuthenticationRequest,
    AuthenticationSuccessResponse,
    ProviderMetadata,
    RpInitiatedLogoutRequest,
    TokenErrorResponse,
    TokenIntrospectionRequest,
    TokenIntrospectionSuccessResponse,
    TokenRequest,
    TokenSuccessResponse,
    UserinfoErrorResponse,
    UserinfoRequest,
    UserinfoSuccessResponse,
)

logger = logging.getLogger(__name__)


def rand_str() -> str:
    return "".join(random.choices(string.ascii_letters + string.digits, k=16))


class DummyOpenidProvider:
    iss = "https://provider.example.com/openid-connect"
    endpoints = {
        "authorization": f"{iss}/client_auth",
        "token": f"{iss}/token",
        "userinfo": f"{iss}/userinfo",
        "end-session": f"{iss}/end-session",
        "token-introspection": f"{iss}/token-introspection",
    }

    test_client_id = "test-client"
    test_client_secret = "foobar123"
    test_client_redirect_uri = "https://app.example.com/auth/callback"

    cheat_code = "master access code which can always be exchanged for tokens"
    cheat_token = "master access token which can always be used for everything"

    _valid_auth_codes: List[str]
    _valid_access_tokens: List[str]

    def __init__(self, requests_mock: responses.RequestsMock):
        self._valid_auth_codes = []
        self._valid_access_tokens = []
        self.setup_authorization_endpoint(requests_mock)
        self.setup_token_endpoint(requests_mock)
        self.setup_userinfo_endpoint(requests_mock)
        self.setup_end_session_endpoint(requests_mock)
        self.setup_token_introspection_endpoint(requests_mock)

    def setup_authorization_endpoint(self, requests_mock: responses.RequestsMock):
        def callback(request: PreparedRequest) -> Tuple[int, Dict[str, str], str]:
            request_url = furl(request.url)
            auth_request = AuthenticationRequest.parse_x_www_form_urlencoded(
                request_url.query.encode()
            )

            code = rand_str()
            self._valid_auth_codes.append(code)
            response = AuthenticationSuccessResponse(
                code=code, state=auth_request.state
            )

            return 301, {"Location": response.encode_url(auth_request.redirect_uri)}, ""

        requests_mock.add_callback(
            method=responses.GET,
            url=self.endpoints["authorization"],
            callback=callback,
        )

    def setup_token_endpoint(self, requests_mock: responses.RequestsMock):
        def callback(request: PreparedRequest) -> Tuple[int, Dict[str, str], str]:
            request_msg = TokenRequest.parse_x_www_form_urlencoded(request.body)
            access_token = rand_str()
            self._valid_access_tokens.append(access_token)
            if request_msg.code == self.cheat_code:
                return (
                    200,
                    {},
                    TokenSuccessResponse(
                        access_token=access_token,
                        token_type="Bearer",
                        id_token=rand_str(),
                    ).json(),
                )
            elif request_msg.code in self._valid_auth_codes:
                self._valid_auth_codes.remove(request_msg.code)
                return (
                    200,
                    {},
                    TokenSuccessResponse(
                        access_token=access_token,
                        token_type="Bearer",
                        id_token=rand_str(),
                    ).json(),
                )
            else:
                return (
                    400,
                    {},
                    TokenErrorResponse(
                        error=TokenErrorResponse.ErrorType.invalid_grant,
                        error_description="token is invalid",
                    ).json(),
                )

        requests_mock.add_callback(
            method=responses.POST,
            url=self.endpoints["token"],
            callback=callback,
            content_type="application/json",
        )

    def setup_userinfo_endpoint(self, requests_mock: responses.RequestsMock):
        def callback(request: PreparedRequest) -> Tuple[int, Dict[str, str], str]:
            _request_msg = UserinfoRequest.parse_x_www_form_urlencoded(request.body)
            if request.headers["Authorization"] == f"Bearer {self.cheat_token}" or any(
                request.headers["Authorization"] == f"Bearer {token}"
                for token in self._valid_access_tokens
            ):
                return (
                    200,
                    {},
                    UserinfoSuccessResponse(sub="abc123", username="test-user").json(),
                )
            else:
                return (
                    400,
                    {},
                    UserinfoErrorResponse(
                        error="invalid_token",
                        error_description="the access token was not valid",
                    ).json(),
                )

        requests_mock.add_callback(
            method=responses.GET,
            url=self.endpoints["userinfo"],
            callback=callback,
            content_type="application/json",
        )

    def setup_end_session_endpoint(self, requests_mock: responses.RequestsMock):
        def callback(request: PreparedRequest) -> Tuple[int, Dict[str, str], str]:
            request_msg = RpInitiatedLogoutRequest.parse_x_www_form_urlencoded(
                request.url.split("?")[1]
            )
            if request_msg.post_logout_redirect_uri is not None:
                return (302, {"Location": request_msg.post_logout_redirect_uri}, "")

            return 200, {}, "Logged out"

        requests_mock.add_callback(
            method=responses.GET,
            url=self.endpoints["end-session"],
            callback=callback,
        )

    def setup_token_introspection_endpoint(self, requests_mock: responses.RequestsMock):
        def callback(request: PreparedRequest) -> Tuple[int, Dict[str, str], str]:
            _request_msg = TokenIntrospectionRequest.parse_x_www_form_urlencoded(
                request.body
            )
            return (
                200,
                {},
                TokenIntrospectionSuccessResponse(
                    active=True,
                    scope="openid",
                    client_id=self.test_client_id,
                    username="test-user",
                    sub="abc123",
                ).json(),
            )

        requests_mock.add_callback(
            method=responses.POST,
            url=self.endpoints["token-introspection"],
            callback=callback,
            content_type="applictaion/json",
        )


@pytest.fixture
def dummy_openid_provider(
    mocked_responses: responses.RequestsMock,
) -> DummyOpenidProvider:
    yield DummyOpenidProvider(mocked_responses)


class DummyUserAgent(requests.Session):
    def naviagte_to(self, url: str) -> requests.Response:
        """
        Mimic navigating to the given URL
        """
        return self.get(url, allow_redirects=True)

    def login_to_dummy_provider(
        self, url: str, username: str = "test", password: str = "foobar123"
    ) -> str:
        """
        Login to the dummy provider by posting the given credentials.
        The credentials are not intended to be real credentials because the provider usually accepts all credentials.

        This is intended to be used against the dummy provider that is defined in this repositories `tests/dummy_provider`.

        :param url: The url which currently presents the login form.
            The user agent is usually redirected here and asked to log in.
        :param username: The username to use during login.
        :param password: The password to use during login.

        :returns: The absolute path on the app server which should now be called to complete the login
        """
        # perform login
        response = self.post(
            url,
            data={
                "prompt": "login",
                "login": username,
                "password": password,
            },
        )

        # give consent
        response = self.post(
            response.url,
            allow_redirects=False,
            data={
                "prompt": "consent",
            },
        )
        while response.headers["Location"].startswith("http://localhost:300"):
            response = self.get(response.headers["Location"], allow_redirects=False)

        assert response.is_redirect, "Consent response did not return a redirect"
        return "/" + response.headers["Location"].split("/", 3)[3]


@pytest.fixture
def dummy_ua(monkeypatch) -> DummyUserAgent:
    yield DummyUserAgent()


class DummyAppServer:
    callback_url = "https://app.example.com/auth/callback"

    def __init__(self, requests_mock: responses.RequestsMock):
        requests_mock.get(
            url=self.callback_url,
            status=200,
            body="dumm app server callback response",
        )


@pytest.fixture
def dummy_app_server(
    mocked_responses: responses.RequestsMock, dummy_openid_provider: DummyOpenidProvider
) -> DummyAppServer:
    return DummyAppServer(mocked_responses)


@pytest.fixture
def mocked_responses():
    """
    A context manager that mocks and de-mocks http request responses
    """
    with responses.RequestsMock(assert_all_requests_are_fired=False) as mock:
        yield mock


@pytest.fixture
def mock_known_provider_configs(mocked_responses: responses.RequestsMock):
    """
    Mock requests to known OpenID provider config URLs to return static content
    """
    # invalid example.com
    mocked_responses.get(
        url="https://example.com/not-found/.well-known/openid-configuration",
        content_type="text/plain",
        body="Issuer config not found",
        status=404,
    )
    mocked_responses.get(
        url="https://example.com/invalid-json-document/.well-known/openid-configuration",
        content_type="application/json",
        body="{ 'hello': 'world' }",
    )
    mocked_responses.get(
        url="https://example.com/invalid-json-syntax/.well-known/openid-configuration",
        content_type="application/json",
        body="{ 'hello }",
    )

    # valid example.com
    mocked_responses.get(
        url="https://provider.example.com/openid-connect/.well-known/openid-configuration",
        json=ProviderMetadata(
            issuer="https://provider.example.com/openid-connect",
            authorization_endpoint="https://provider.example.com/openid-connect/client_auth",
            token_endpoint="https://provider.example.com/openid-connect/token",
            jwks_uri="https://provider.example.com/openid-connect/jwks",
            userinfo_endpoint="https://provider.example.com/openid-connect/userinfo",
            end_session_endpoint="https://provider.example.com/openid-connect/end-session",
            introspection_endpoint="https://provider.example.com/openid-connect/token-introspection",
            subject_types_supported=["public"],
            id_token_signing_alg_values_supported=["RS256"],
        ).dict(exclude_defaults=True),
    )
    mocked_responses.get(
        url="https://provider.example.com/openid-connect/jwks",
        json={"keys": []},
    )

    # mafiasi identity
    mocked_responses.get(
        url="https://identity.mafiasi.de/auth/realms/simple_openid_test/.well-known/openid-configuration",
        content_type="application/json",
        body='{"issuer":"https://identity.mafiasi.de/auth/realms/simple_openid_test","authorization_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/auth","token_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/token","introspection_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/token/introspect","userinfo_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/userinfo","end_session_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/logout","frontchannel_logout_session_supported":true,"frontchannel_logout_supported":true,"jwks_uri":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/certs","check_session_iframe":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/login-status-iframe.html","grant_types_supported":["authorization_code","implicit","refresh_token","password","client_credentials","urn:ietf:params:oauth:grant-type:device_code","urn:openid:params:grant-type:ciba"],"response_types_supported":["code","none","id_token","token","id_token token","code id_token","code token","code id_token token"],"subject_types_supported":["public","pairwise"],"id_token_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512"],"id_token_encryption_alg_values_supported":["RSA-OAEP","RSA-OAEP-256","RSA1_5"],"id_token_encryption_enc_values_supported":["A256GCM","A192GCM","A128GCM","A128CBC-HS256","A192CBC-HS384","A256CBC-HS512"],"userinfo_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512","none"],"request_object_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512","none"],"request_object_encryption_alg_values_supported":["RSA-OAEP","RSA-OAEP-256","RSA1_5"],"request_object_encryption_enc_values_supported":["A256GCM","A192GCM","A128GCM","A128CBC-HS256","A192CBC-HS384","A256CBC-HS512"],"response_modes_supported":["query","fragment","form_post","query.jwt","fragment.jwt","form_post.jwt","jwt"],"registration_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/clients-registrations/openid-connect","token_endpoint_auth_methods_supported":["private_key_jwt","client_secret_basic","client_secret_post","tls_client_auth","client_secret_jwt"],"token_endpoint_auth_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512"],"introspection_endpoint_auth_methods_supported":["private_key_jwt","client_secret_basic","client_secret_post","tls_client_auth","client_secret_jwt"],"introspection_endpoint_auth_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512"],"authorization_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512"],"authorization_encryption_alg_values_supported":["RSA-OAEP","RSA-OAEP-256","RSA1_5"],"authorization_encryption_enc_values_supported":["A256GCM","A192GCM","A128GCM","A128CBC-HS256","A192CBC-HS384","A256CBC-HS512"],"claims_supported":["aud","sub","iss","auth_time","name","given_name","family_name","preferred_username","email","acr"],"claim_types_supported":["normal"],"claims_parameter_supported":true,"scopes_supported":["openid","address","web-origins","offline_access","profile","email","roles","phone","microprofile-jwt"],"request_parameter_supported":true,"request_uri_parameter_supported":true,"require_request_uri_registration":true,"code_challenge_methods_supported":["plain","S256"],"tls_client_certificate_bound_access_tokens":true,"revocation_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/revoke","revocation_endpoint_auth_methods_supported":["private_key_jwt","client_secret_basic","client_secret_post","tls_client_auth","client_secret_jwt"],"revocation_endpoint_auth_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512"],"backchannel_logout_supported":true,"backchannel_logout_session_supported":true,"device_authorization_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/auth/device","backchannel_token_delivery_modes_supported":["poll","ping"],"backchannel_authentication_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/ext/ciba/auth","backchannel_authentication_request_signing_alg_values_supported":["PS384","ES384","RS384","ES256","RS256","ES512","PS256","PS512","RS512"],"require_pushed_authorization_requests":false,"pushed_authorization_request_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/ext/par/request","mtls_endpoint_aliases":{"token_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/token","revocation_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/revoke","introspection_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/token/introspect","device_authorization_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/auth/device","registration_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/clients-registrations/openid-connect","userinfo_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/userinfo","pushed_authorization_request_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/ext/par/request","backchannel_authentication_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/ext/ciba/auth"}}',
    )

    # google
    mocked_responses.get(
        "https://accounts.google.com/.well-known/openid-configuration",
        content_type="application/json",
        body='{"issuer":"https://accounts.google.com","authorization_endpoint":"https://accounts.google.com/o/oauth2/v2/auth","device_authorization_endpoint":"https://oauth2.googleapis.com/device/code","token_endpoint":"https://oauth2.googleapis.com/token","userinfo_endpoint":"https://openidconnect.googleapis.com/v1/userinfo","revocation_endpoint":"https://oauth2.googleapis.com/revoke","jwks_uri":"https://www.googleapis.com/oauth2/v3/certs","response_types_supported":["code","token","id_token","code token","code id_token","token id_token","code token id_token","none"],"subject_types_supported":["public"],"id_token_signing_alg_values_supported":["RS256"],"scopes_supported":["openid","email","profile"],"token_endpoint_auth_methods_supported":["client_secret_post","client_secret_basic"],"claims_supported":["aud","email","email_verified","exp","family_name","given_name","iat","iss","locale","name","picture","sub"],"code_challenge_methods_supported":["plain","S256"],"grant_types_supported":["authorization_code","refresh_token","urn:ietf:params:oauth:grant-type:device_code","urn:ietf:params:oauth:grant-type:jwt-bearer"]}',
    )


@pytest.fixture
def mock_known_provider_jwks(mocked_responses: responses.RequestsMock):
    mocked_responses.get(
        url="https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/certs",
        content_type="application/json",
        body='{"keys":[{"kid":"JAehgk0O0uzTEht7KGCPVB_urwfsBGe22phHVDZezeo","kty":"RSA","alg":"RS256","use":"sig","n":"n1t4R6mVlBZggtmEM5HohZYg5QjLtjlRVyAMMM93H4WRPpP2Xqj3RrGwP-LZr6lQhk6w8nvPNGnFujWtXb_dBrhZDBh2z0mCNvfjJMp5Ki5sLT9nCbuZ4NkIhfx7qTzQ--GomQ6uKOSuoE12A_r9UopwuCf_1zrx0MIqN_kMeYU2px5yJW-5qbwPL9hcWcfY_Pdz5SFcQ3kDi6GlXA0PN74Ivhs9baVIv60cwB0TytQXGn__GYFZ-K5qAjIymyIy8cPRy9N27Gs29cWF-C56n1pAxwkcPY1SAQgcrspIhntQbZBeGboUx2pnQ_v8LEqVt9ahh52eAbV36YoQqu1DbQ","e":"AQAB","x5c":["MIICszCCAZsCBgGEZpqrijANBgkqhkiG9w0BAQsFADAdMRswGQYDVQQDDBJzaW1wbGVfb3BlbmlkX3Rlc3QwHhcNMjIxMTExMTIxMDIzWhcNMzIxMTExMTIxMjAzWjAdMRswGQYDVQQDDBJzaW1wbGVfb3BlbmlkX3Rlc3QwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCfW3hHqZWUFmCC2YQzkeiFliDlCMu2OVFXIAwwz3cfhZE+k/ZeqPdGsbA/4tmvqVCGTrDye880acW6Na1dv90GuFkMGHbPSYI29+MkynkqLmwtP2cJu5ng2QiF/HupPND74aiZDq4o5K6gTXYD+v1SinC4J//XOvHQwio3+Qx5hTanHnIlb7mpvA8v2FxZx9j893PlIVxDeQOLoaVcDQ83vgi+Gz1tpUi/rRzAHRPK1Bcaf/8ZgVn4rmoCMjKbIjLxw9HL03bsazb1xYX4LnqfWkDHCRw9jVIBCByuykiGe1BtkF4ZuhTHamdD+/wsSpW31qGHnZ4BtXfpihCq7UNtAgMBAAEwDQYJKoZIhvcNAQELBQADggEBAD3aKq2xiLNPlUcmxwtR1FihkEM+NM7VV+m+74Lf9m/lhNzuM00JZAW67gIz2kO5K4XaU84PCRzt3sELma/xVoBs3iW8AhtldTax17g0NZzpMEExzYi0vyihq8xihbx+8XrVwxQ3JBkb/y3b+otDX3xRxBKSCDuxuigL1bX7UzgJ3oe4wD6m8yN7nrMdqSMbspd8lkceKNPziclXnbKPIUwrnxdWMmB41zYcT3y+zHxqk6EDkf3NU+C3gCoRCQoBKcK4X2dwZdUbv0ginU9kROjNXc2+gP53CYDT4fqOEX/jBuljbYb9ziaQ9/5kJZ9SsazgWNjpf4VvhAsH64i7+dE="],"x5t":"PxLqu4M4uNWaUhbfkvI0tGrkhtQ","x5t#S256":"OWVSYSSNFaLyfxV0Lg6XyVDnB2t6sHwtNbzVB2fNCVc"}]}',
    )


@pytest.fixture
def secrets() -> Mapping[str, str]:
    path = Path(__file__).parent / "secrets.yml"
    with open(path, "r", encoding="UTF-8") as f:
        return yaml.safe_load(f)



"""
JSON-Web-Key handling code
"""

from typing import List

from cryptojwt import JWK, KeyBundle


def fetch_jwks(jwks_uri: str) -> List[JWK]:
    """
    Fetch JSON web keys from the given jwks_uri.
    This uri is part of the provider configuration and used to validate responses and tokens sent by the provider.
    """
    bundle = KeyBundle(source=jwks_uri)
    return bundle.keys()  # type: ignore # because cryptojwk has no typedefs, but we know what this returns



import pytest


@pytest.fixture
def mafiasi_identity_settings(settings, secrets):
    settings.OPENID_ISSUER = (
        "https://identity.mafiasi.de/auth/realms/simple_openid_test"
    )
    settings.OPENID_CLIENT_ID = secrets["mafiasi_identity_confidential_client_id"]
    settings.OPENID_CLIENT_SECRET = secrets[
        "mafiasi_identity_confidential_client_secret"
    ]



"""
View functions which handle openid authentication and their related callbacks
"""

from http import HTTPStatus
from typing import Mapping

from django.conf import settings
from django.contrib.auth import login, logout
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect
from django.shortcuts import resolve_url
from django.template.response import TemplateResponse
from django.views import View

from simple_openid_connect.data import IdToken, TokenSuccessResponse
from simple_openid_connect_django.apps import OpenidAppConfig
from simple_openid_connect_django.models import OpenidUser


def _get_redirect_args(request: HttpRequest) -> Mapping[str, str]:
    """
    Calculate arguments which are appended to the redirect_uri
    """
    if "next" in request.GET.keys():
        return {
            "next": request.GET["next"],
        }
    return {}


class InitLoginView(View):
    """
    The view which handles initiating a login.

    It essentially redirects the user agent to the Openid provider.
    """

    def get(self, request: HttpRequest) -> HttpResponse:
        logout(request)
        client = OpenidAppConfig.get_instance().get_client(request)
        redirect = client.authorization_code_flow.start_authentication(
            additional_redirect_args=_get_redirect_args(request),
        )
        return HttpResponseRedirect(redirect)


class LoginCallbackView(View):
    """
    The view which handles login callbacks.

    It handles an authentication response from the Openid provider that is encoded in the current url by either logging
    the user in or rendering the error.

    Error rendering can be customized by overwriting the template *simple_openid_connect_django/login_failed.html* which
    receives the context `token_response` of type :class:`TokenErrorResponse <simple_openid_connect.data.TokenErrorResponse>`.
    """

    def get(self, request: HttpRequest) -> HttpResponse:
        client = OpenidAppConfig.get_instance().get_client(request)

        token_response = client.authorization_code_flow.handle_authentication_result(
            current_url=request.get_full_path(),
            additional_redirect_args=_get_redirect_args(request),
        )
        if not isinstance(token_response, TokenSuccessResponse):
            return TemplateResponse(
                request,
                "simple_openid_connect_django/login_failed.html",
                {
                    "token_response": token_response,
                },
                status=HTTPStatus.UNAUTHORIZED,
            )

        id_token = IdToken.parse_jwt(token_response.id_token, client.provider_keys)
        id_token.validate_extern(
            client.provider_config.issuer, client.client_auth.client_id
        )

        user = OpenidUser.objects.get_or_create_from_id_token(id_token)
        user.update_session(token_response)
        login(request, user.user)

        # redirect to the next get parameter if present, otherwise to the configured default
        if "next" in request.GET.keys():
            return HttpResponseRedirect(redirect_to=request.GET["next"])
        else:
            return HttpResponseRedirect(
                redirect_to=resolve_url(settings.LOGIN_REDIRECT_URL)
            )


class LogoutView(View):
    """
    The view which handles logging a user out.
    """

    def get(self, request: HttpRequest) -> HttpResponse:
        client = OpenidAppConfig.get_instance().get_client(request)
        return HttpResponseRedirect(client.initiate_logout())



"""
View-function decorators
"""

import logging
from functools import wraps
from http import HTTPStatus
from typing import Any, Callable, TypeVar, Union

from django.http import HttpRequest, HttpResponse, JsonResponse

from simple_openid_connect.data import TokenIntrospectionErrorResponse
from simple_openid_connect.utils import is_application_json
from simple_openid_connect_django.apps import OpenidAppConfig

logger = logging.getLogger(__name__)

View_Return = TypeVar("View_Return", bound=HttpResponse)


def access_token_required(
    *, required_scopes: str = "openid"
) -> Callable[..., Union[HttpResponse, View_Return]]:
    """
    Decorator for views that checks that the request is authenticated using a valid access token, early-returning an
    appropriate http error response if necessary.

    :param required_scopes: Scopes to which the access token needs to have access

    :raises UnsupportedByProviderError: If the provider does not support token introspection.
    """

    def actual_decorator(
        view_func: Callable[..., View_Return]
    ) -> Callable[..., View_Return]:
        @wraps(view_func)
        def wrapped_view(
            request: HttpRequest, *args: Any, **kwargs: Any
        ) -> Union[HttpResponse, View_Return]:
            # verify that an Authorization Header of type Bearer is present
            if "Authorization" not in request.headers.keys() or not request.headers[
                "Authorization"
            ].startswith("Bearer "):
                return HttpResponse(
                    status=HTTPStatus.UNAUTHORIZED,
                    headers={"WWW-Authenticate": "Bearer"},
                )

            # introspect passed token
            token = request.headers["Authorization"].split(" ", 1)[1]
            oidc_client = OpenidAppConfig.get_instance().get_client(request)
            result = oidc_client.introspect_token(token)

            if isinstance(result, TokenIntrospectionErrorResponse):
                logger.critical("could not introspect token for validity: %s", result)
                return HttpResponse(status=HTTPStatus.INTERNAL_SERVER_ERROR)
            if result.scope is None:
                logger.critical(
                    "could not determine access because token introspection did not return token scopes"
                )
                return HttpResponse(status=HTTPStatus.INTERNAL_SERVER_ERROR)

            # directly return a http response if the token is not valid
            if not result.active or any(
                scope not in result.scope.split(" ")
                for scope in required_scopes.split(" ")
            ):
                if "Accept" in request.headers.keys() and is_application_json(
                    request.headers["Accept"]
                ):
                    return JsonResponse(
                        status=HTTPStatus.UNAUTHORIZED,
                        headers={"WWW-Authenticate": "Bearer"},
                        data={
                            "error": "invalid_token",
                            "error_description": "the used access token is not valid or does not grant enough access",
                        },
                    )
                else:
                    return HttpResponse(
                        status=HTTPStatus.UNAUTHORIZED,
                        content="the used access token is not valid or does not grant enough access",
                        headers={"WWW-Authenticate": "Bearer"},
                    )

            # execute the decorated view function
            return view_func(request, *args, **kwargs)

        return wrapped_view  # type: ignore

    return actual_decorator  # type: ignore



"""
Django AppConfig for this app
"""

import logging
from typing import Any, Callable, Optional, Union

from django.apps import AppConfig, apps
from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest
from django.shortcuts import resolve_url
from django.utils.module_loading import import_string
from pydantic import BaseModel

from simple_openid_connect.client import OpenidClient
from simple_openid_connect.data import IdToken

logger = logging.getLogger(__name__)


class SettingsModel(BaseModel):
    """
    A pydantic model used to validate django settings
    """

    OPENID_ISSUER: str
    OPENID_CLIENT_ID: str
    OPENID_CLIENT_SECRET: Optional[str]
    OPENID_SCOPE: str = "openid"
    OPENID_REDIRECT_URI: str = "simple_openid_connect_django:login-callback"
    OPENID_BASE_URI: Optional[str]
    OPENID_CREATE_USER_FUNC = (
        "simple_openid_connect_django.user_mapping.create_user_from_token"
    )
    OPENID_UPDATE_USER_FUNC = (
        "simple_openid_connect_django.user_mapping.update_user_from_token"
    )

    class Config:
        orm_mode = True


class OpenidAppConfig(AppConfig):
    name = "simple_openid_connect_django"

    def ready(self) -> None:
        """
        Called when django starts.

        Performs settings validation and raises ImproperlyConfigured if necessary.
        """
        super().ready()

        # assert that are settings are as required
        try:
            _ = self.safe_settings
            _ = self.create_user_func
            _ = self.update_user_func
        except Exception as e:
            raise ImproperlyConfigured(
                f"django settings are invalid for openid usage: {e}"
            ) from e

    @classmethod
    def get_instance(cls) -> "OpenidAppConfig":
        """
        Retrieve the currently used instance from django's app registry
        """
        instance = apps.get_app_config(cls.name)
        assert isinstance(instance, OpenidAppConfig)
        return instance

    @property
    def safe_settings(self) -> SettingsModel:
        """
        type-validated version of django settings
        """
        return SettingsModel.from_orm(settings)

    @property
    def create_user_func(self) -> Callable[[IdToken], Any]:
        """
        The function which is configured via django settings and which creates new users based on id tokens.
        """
        return import_string(self.safe_settings.OPENID_CREATE_USER_FUNC)  # type: ignore

    @property
    def update_user_func(self) -> Callable[[Any, IdToken], None]:
        """
        The function which is configured via django settings and which updates user objects based on id tokens.
        """
        return import_string(self.safe_settings.OPENID_UPDATE_USER_FUNC)  # type: ignore

    def get_client(
        self, own_base_uri: Union[HttpRequest, str, None] = None
    ) -> OpenidClient:
        """
        Get an `OpenidClient` instance that is appropriate for usage in django.

        It is automatically configured via django settings.

        :param own_base_uri: The base url of this application which will be used to construct a redirect_uri back to it.
            Can also be the current request in which case `{scheme}://{host}` of it will be used as the base url.
            If this parameter is not given, only the `OPENID_BASE_URI` setting is used.
            In any case, if the `OPENID_BASE_URI` setting is set, it will be used instead.

        :raises ImproperlyConfigured: when no *own_base_uri* is given and the `OPENID_BASE_URI` is also None
        """
        # determine base_uri of this app
        if self.safe_settings.OPENID_BASE_URI is not None:
            own_base_uri = self.safe_settings.OPENID_BASE_URI
        else:
            if own_base_uri is None:
                raise ImproperlyConfigured(
                    "either a value for own_base_uri must be given or the django setting OPENID_BASE_URI must be filled"
                )
            elif isinstance(own_base_uri, HttpRequest):
                own_base_uri = f"{own_base_uri.scheme}://{own_base_uri.get_host()}"

        # use a cached client instance if one exists or create a new one if not
        client = cache.get("openid_client")  # type: OpenidClient
        if client is None:
            redirect_uri = resolve_url(self.safe_settings.OPENID_REDIRECT_URI)
            client = OpenidClient.from_issuer_url(
                url=self.safe_settings.OPENID_ISSUER,
                authentication_redirect_uri=f"{own_base_uri}{redirect_uri}",
                client_id=self.safe_settings.OPENID_CLIENT_ID,
                client_secret=self.safe_settings.OPENID_CLIENT_SECRET,
                scope=self.safe_settings.OPENID_SCOPE,
            )
            cache.set("openid_client", client)

        return client



import pickle
import re

from simple_openid_connect.client import OpenidClient
from simple_openid_connect.data import (
    RpInitiatedLogoutRequest,
    TokenIntrospectionSuccessResponse,
    UserinfoSuccessResponse,
)


def test_full_authorization_code_flow(
    mock_known_provider_configs, dummy_openid_provider, dummy_ua, dummy_app_server
):
    # act
    client = OpenidClient.from_issuer_url(
        url="https://provider.example.com/openid-connect",
        authentication_redirect_uri=dummy_app_server.callback_url,
        client_id=dummy_openid_provider.test_client_id,
        client_secret=dummy_openid_provider.test_client_secret,
    )
    response = dummy_ua.naviagte_to(
        client.authorization_code_flow.start_authentication()
    )
    result = client.authorization_code_flow.handle_authentication_result(response.url)

    # assert
    assert result.access_token
    assert result.id_token


def test_additional_redirect_args(
    mocked_responses,
    mock_known_provider_configs,
    dummy_openid_provider,
    dummy_ua,
    dummy_app_server,
):
    # arrange
    client = OpenidClient.from_issuer_url(
        url="https://provider.example.com/openid-connect",
        authentication_redirect_uri=dummy_app_server.callback_url,
        client_id=dummy_openid_provider.test_client_id,
        client_secret=dummy_openid_provider.test_client_secret,
    )

    # act
    auth_response = dummy_ua.naviagte_to(
        client.authorization_code_flow.start_authentication(
            additional_redirect_args={"foo": "bar"}
        )
    )
    _code_response = client.authorization_code_flow.handle_authentication_result(
        auth_response.url, additional_redirect_args={"foo": "bar"}
    )

    # assert
    assert "foo=bar" in auth_response.url
    assert any(
        call.request.body and re.search(r"redirect_uri=.*foo%3Dbar", call.request.body)
        for call in mocked_responses.calls
    )


def test_client_type(mock_known_provider_configs):
    # arrange
    public_client = OpenidClient.from_issuer_url(
        url="https://provider.example.com/openid-connect",
        authentication_redirect_uri="",
        client_id="test",
    )
    confidential_client = OpenidClient.from_issuer_url(
        url="https://provider.example.com/openid-connect",
        authentication_redirect_uri="",
        client_id="test",
        client_secret="test",
    )

    # assert
    assert public_client.client_type == "public"
    assert confidential_client.client_type == "confidential"


def test_fetch_userinfo(
    mock_known_provider_configs, dummy_openid_provider, dummy_app_server, dummy_ua
):
    # arrange
    client = OpenidClient.from_issuer_url(
        url="https://provider.example.com/openid-connect",
        authentication_redirect_uri=dummy_app_server.callback_url,
        client_id=dummy_openid_provider.test_client_id,
        client_secret=dummy_openid_provider.test_client_secret,
    )

    # act
    response = client.fetch_userinfo(dummy_openid_provider.cheat_token)

    # assert
    assert isinstance(response, UserinfoSuccessResponse)
    assert response.sub == "abc123"
    assert response.username == "test-user"


def test_rp_initiated_logout(
    mock_known_provider_configs, dummy_openid_provider, dummy_app_server, dummy_ua
):
    # arrange
    client = OpenidClient.from_issuer_url(
        url="https://provider.example.com/openid-connect",
        authentication_redirect_uri=dummy_app_server.callback_url,
        client_id=dummy_openid_provider.test_client_id,
        client_secret=dummy_openid_provider.test_client_secret,
    )

    # act
    plain_url = client.initiate_logout()
    advanced_url = client.initiate_logout(
        RpInitiatedLogoutRequest(post_logout_redirect_uri=dummy_app_server.callback_url)
    )
    nav_response = dummy_ua.naviagte_to(advanced_url)

    # assert
    assert plain_url == "https://provider.example.com/openid-connect/end-session"
    assert (
        advanced_url
        == "https://provider.example.com/openid-connect/end-session?post_logout_redirect_uri=https%3A%2F%2Fapp.example.com%2Fauth%2Fcallback"
    )
    assert nav_response.url == dummy_app_server.callback_url


def test_token_introspection(mock_known_provider_configs, dummy_openid_provider):
    # arrange
    client = OpenidClient.from_issuer_url(
        url="https://provider.example.com/openid-connect",
        authentication_redirect_uri="",
        client_id=dummy_openid_provider.test_client_id,
        client_secret=dummy_openid_provider.test_client_secret,
    )

    # act
    response = client.introspect_token(dummy_openid_provider.cheat_token)

    # assert
    assert isinstance(response, TokenIntrospectionSuccessResponse)
    assert response.active


def test_pickling(mock_known_provider_configs, dummy_openid_provider):
    # arrange
    client = OpenidClient.from_issuer_url(
        url="https://provider.example.com/openid-connect",
        authentication_redirect_uri="",
        client_id=dummy_openid_provider.test_client_id,
        client_secret=dummy_openid_provider.test_client_secret,
    )

    # act (assert not throwing)
    enc = pickle.dumps(client)
    _ = pickle.loads(enc)



import pytest

from simple_openid_connect import discovery
from simple_openid_connect.exceptions import OpenidProtocolError


def test_mafiasi_identity(mock_known_provider_configs):
    discovery.discover_configuration_from_issuer(
        "https://identity.mafiasi.de/auth/realms/simple_openid_test"
    )


def test_google(mock_known_provider_configs):
    discovery.discover_configuration_from_issuer("https://accounts.google.com")


def test_errors(mock_known_provider_configs):
    with pytest.raises(OpenidProtocolError):
        discovery.discover_configuration_from_issuer("https://example.com/not-found")
    with pytest.raises(OpenidProtocolError):
        discovery.discover_configuration_from_issuer(
            "https://example.com/invalid-json-document"
        )
    with pytest.raises(OpenidProtocolError):
        discovery.discover_configuration_from_issuer(
            "https://example.com/invalid-json-syntax"
        )



import logging
from http import HTTPStatus
from typing import Mapping, Tuple

import pytest
from furl import furl

from simple_openid_connect.client import OpenidClient
from simple_openid_connect.data import (
    IdToken,
    TokenIntrospectionErrorResponse,
    TokenIntrospectionSuccessResponse,
    UserinfoSuccessResponse,
)
from simple_openid_connect.flows.authorization_code_flow import TokenSuccessResponse

logger = logging.getLogger(__name__)


@pytest.mark.interactive()
def test_public_client(real_app_server, secrets):
    # arrange
    oidc_client = OpenidClient.from_issuer_url(
        "https://identity.mafiasi.de/auth/realms/simple_openid_test",
        real_app_server.callback_url,
        client_id=secrets["mafiasi_identity_public_client_id"],
    )
    token_response: TokenSuccessResponse

    def on_login(_url: furl) -> Tuple[int, Mapping[str, str], str]:
        url = oidc_client.authorization_code_flow.start_authentication()
        return (
            HTTPStatus.FOUND,
            {
                "Location": url,
            },
            f"Go to {url}",
        )

    def on_login_callback(url: furl) -> Tuple[int, Mapping[str, str], str]:
        nonlocal token_response
        response = oidc_client.authorization_code_flow.handle_authentication_result(
            str(url)
        )
        assert isinstance(response, TokenSuccessResponse)
        token_response = response
        real_app_server.done()
        return HTTPStatus.OK, {}, f"Success. You can close this tab."

    # act (login)
    logger.info(f"Visit {real_app_server.login_url}")
    real_app_server.serve_until_done(on_login, on_login_callback)

    # act (decode id token)
    id_token = oidc_client.decode_id_token(token_response.id_token)
    assert isinstance(id_token, IdToken)
    id_token.validate_extern(
        oidc_client.provider_config.issuer, oidc_client.client_auth.client_id
    )

    # act (get userinfo)
    response = oidc_client.fetch_userinfo(token_response.access_token)
    assert isinstance(response, UserinfoSuccessResponse)
    assert response.sub != ""

    # act (introspect access token). This fails because te public client cannot authenticate itself
    response = oidc_client.introspect_token(token_response.access_token, "access_token")
    assert isinstance(response, TokenIntrospectionErrorResponse)
    assert response.error == "invalid_request"
    assert response.error_description == "Authentication failed."

    # act (refresh tokens)
    response = oidc_client.exchange_refresh_token(token_response.refresh_token)
    assert isinstance(response, TokenSuccessResponse)


@pytest.mark.interactive()
def test_confidential_client(real_app_server, secrets):
    # arrange
    oidc_client = OpenidClient.from_issuer_url(
        "https://identity.mafiasi.de/auth/realms/simple_openid_test",
        real_app_server.callback_url,
        client_id=secrets["mafiasi_identity_confidential_client_id"],
        client_secret=secrets["mafiasi_identity_confidential_client_secret"],
    )
    token_response: TokenSuccessResponse

    def on_login(_url: furl) -> Tuple[int, Mapping[str, str], str]:
        url = oidc_client.authorization_code_flow.start_authentication()
        return (
            HTTPStatus.FOUND,
            {
                "Location": url,
            },
            f"Go to {url}",
        )

    def on_login_callback(url: furl) -> Tuple[int, Mapping[str, str], str]:
        nonlocal token_response
        response = oidc_client.authorization_code_flow.handle_authentication_result(
            str(url)
        )
        assert isinstance(response, TokenSuccessResponse)
        token_response = response
        real_app_server.done()
        return HTTPStatus.OK, {}, f"Success. You can close this tab."

    # act (login)
    logger.info(f"Visit {real_app_server.login_url}")
    real_app_server.serve_until_done(on_login, on_login_callback)

    # act (decode id token)
    id_token = oidc_client.decode_id_token(token_response.id_token)
    assert isinstance(id_token, IdToken)
    id_token.validate_extern(
        oidc_client.provider_config.issuer, oidc_client.client_auth.client_id
    )

    # act (get userinfo)
    response = oidc_client.fetch_userinfo(token_response.access_token)
    assert isinstance(response, UserinfoSuccessResponse)
    assert response.sub != ""

    # act (introspect access token)
    response = oidc_client.introspect_token(token_response.access_token, "access_token")
    assert isinstance(response, TokenIntrospectionSuccessResponse)

    # act (introspect access token)
    response = oidc_client.introspect_token(
        token_response.refresh_token, "refresh_token"
    )
    assert isinstance(response, TokenIntrospectionSuccessResponse)

    # act (refresh tokens)
    response = oidc_client.exchange_refresh_token(token_response.refresh_token)
    assert isinstance(response, TokenSuccessResponse)



"""simple_openid_connect_django URL Configuration

These urls should be included into your project wherever you like via::

    urlpatterns = [
        ...
        path("auth/openid/", include("simple_openid_connect_django.urls")),
    ]

"""
from django.urls import path

from simple_openid_connect_django import views

app_name = "simple_openid_connect_django"
urlpatterns = [
    path("login/", views.InitLoginView.as_view(), name="login"),
    path("login-callback/", views.LoginCallbackView.as_view(), name="login-callback"),
    path("logout/", views.LogoutView.as_view(), name="logout"),
]



from base64 import b64encode

import pytest
import requests
from responses import RequestsMock

from simple_openid_connect.client_authentication import ClientSecretBasicAuth, NoneAuth


@pytest.fixture
def mock_empty_response(mocked_responses: RequestsMock):
    mocked_responses.get(url="https://example.com")


def test_none_auth(mock_empty_response):
    # arrange
    auth = NoneAuth("test-id")

    # act
    response = requests.get("https://example.com", auth=auth)

    # assert
    assert "Authorization" not in response.request.headers.keys()


def test_client_secret_basic_auth(mock_empty_response):
    # arrange
    auth = ClientSecretBasicAuth("test-id", "foobar123")

    # act
    response = requests.get("https://example.com", auth=auth)

    # assert
    assert (
        response.request.headers["Authorization"]
        == f"Basic {b64encode(b':'.join(['test-id'.encode('ASCII'), 'foobar123'.encode('ASCII')])).decode('ASCII')}"
    )



from django.contrib.auth.decorators import login_required
from django.http import HttpRequest, HttpResponse

from simple_openid_connect_django.decorators import access_token_required


@login_required
def default_after_login(_request: HttpRequest) -> HttpResponse:
    return HttpResponse(content="default login redirect view")


@login_required
def test_protected_view(request: HttpRequest) -> HttpResponse:
    return HttpResponse(content=f"hello user {request.user.openid.sub}")


@access_token_required()
def test_access_token_view(_request: HttpRequest) -> HttpResponse:
    return HttpResponse(content="access granted")
